import 'package:flutter/material.dart';
import 'package:mock_data/mock_data.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:phrase/src/cache/phrase_disk_cache.dart';
import 'package:phrase/src/phrase_api.dart';
import 'package:phrase/src/phrase_chef.dart';
import 'package:phrase/src/phrase_impl.dart';
import 'package:phrase/src/phrase_prefs.dart';
import 'package:test/test.dart';

import '../_helper/arb.dart';
import 'phrase_impl_test.mocks.dart';

@GenerateMocks([
  PhraseDiskCache,
  PhrasePrefs,
  PhraseApi,
  PhraseChef,
])
void main() {
  group('PhraseImpl', () {
    final diskCacheMock = MockPhraseDiskCache();
    final prefsMock = MockPhrasePrefs();
    final apiMock = MockPhraseApi();
    final chefMock = MockPhraseChef();

    final distId = mockUUID();
    final envId = mockUUID();

    setUp(() {
      reset(diskCacheMock);
      reset(prefsMock);
      reset(apiMock);
      reset(chefMock);
    });

    group('loadFromApi', () {
      final phraseImpl = PhraseImpl(
        distId,
        envId,
        true,
        null,
        PhraseHost.eu,
        diskCache: diskCacheMock,
        prefs: prefsMock,
        api: apiMock,
        chef: chefMock,
      );

      final uuid = mockUUID();
      const locale = 'de';
      const version = '1';
      final lastUpdate = DateTime.now().toIso8601String();
      final arb = loadArb('test.arb');
      final localeHash = PhraseImpl.getLocaleHash(distId, envId, locale);

      setUp(() {
        when(prefsMock.getUUID()).thenAnswer((realInvocation) async => uuid);
        when(prefsMock.getVersion(any))
            .thenAnswer((realInvocation) async => version);
        when(prefsMock.getLastUpdate())
            .thenAnswer((realInvocation) async => lastUpdate);
      });

      test('loads modified translations', () async {
        final apiResult = PhraseApiResult.success(version, arb);

        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);

        await phraseImpl.loadFromApi(locale, true);

        verify(diskCacheMock.write(localeHash, arb));
        verify(prefsMock.getUUID());
        verify(prefsMock.getLastUpdate());
        verify(prefsMock.getVersion(localeHash));
        verify(prefsMock.setLastUpdate());
        verify(prefsMock.setVersion(localeHash, version));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.eu,
          locale,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
        ));

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });

      test('translations not modified', () async {
        final apiResult = PhraseApiResult.notModified();

        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);

        await phraseImpl.loadFromApi(locale, true);

        verify(prefsMock.getUUID());
        verify(prefsMock.getLastUpdate());
        verify(prefsMock.getVersion(localeHash));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.eu,
          locale,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
        ));

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });

      test('loads from us host when set', () async {
        final phraseImpl = PhraseImpl(
          distId,
          envId,
          true,
          null,
          PhraseHost.us,
          diskCache: diskCacheMock,
          prefs: prefsMock,
          api: apiMock,
          chef: chefMock,
        );
        final apiResult = PhraseApiResult.success(version, arb);

        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);

        await phraseImpl.loadFromApi(locale, true);

        verify(diskCacheMock.write(localeHash, arb));
        verify(prefsMock.getUUID());
        verify(prefsMock.getLastUpdate());
        verify(prefsMock.getVersion(localeHash));
        verify(prefsMock.setLastUpdate());
        verify(prefsMock.setVersion(localeHash, version));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.us,
          locale,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
        ));

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });

      test('fails to load translation', () async {
        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any))
            .thenThrow(Exception());

        await phraseImpl.loadFromApi(locale, true);
        verify(prefsMock.getUUID());
        verify(prefsMock.getLastUpdate());
        verify(prefsMock.getVersion(localeHash));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.eu,
          locale,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
        ));

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });
    });

    group('getText', () {
      final phraseImpl = PhraseImpl(
        distId,
        envId,
        true,
        null,
        PhraseHost.eu,
        diskCache: diskCacheMock,
        prefs: prefsMock,
        api: apiMock,
        chef: chefMock,
      );

      final uuid = mockUUID();
      const locale = 'de';
      const version = '1';
      final lastUpdate = DateTime.now().toIso8601String();
      final arbFileContents = loadArb('test.arb');

      setUp(() async {
        when(prefsMock.getUUID()).thenAnswer((realInvocation) async => uuid);
        when(prefsMock.getVersion(any))
            .thenAnswer((realInvocation) async => version);
        when(prefsMock.getLastUpdate())
            .thenAnswer((realInvocation) async => lastUpdate);

        final apiResult = PhraseApiResult.success(version, arbFileContents);
        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);
        await phraseImpl.loadFromApi(locale, true);

        reset(diskCacheMock);
        reset(prefsMock);
        reset(apiMock);
        reset(chefMock);
      });

      test('success', () async {
        const key = 'text';
        when(chefMock.cook(any, any, any)).thenReturn('expected');

        final result = phraseImpl.getText(locale, key);

        expect(result, 'expected');

        verify(chefMock.cook(locale, any, key));
        verifyNoMoreInteractions(chefMock);
        verifyZeroInteractions(diskCacheMock);
        verifyZeroInteractions(prefsMock);
        verifyZeroInteractions(apiMock);
      });

      test('translation not found', () {
        const key = 'text';
        when(chefMock.cook(any, any, any)).thenReturn(null);
        final result = phraseImpl.getText(locale, key);

        expect(result, null);

        verify(chefMock.cook(locale, any, key));
        verifyNoMoreInteractions(chefMock);
        verifyZeroInteractions(diskCacheMock);
        verifyZeroInteractions(prefsMock);
        verifyZeroInteractions(apiMock);
      });
    });

    group('onLocale', () {
      const localeCode = 'de';
      const locale = Locale(localeCode);
      final fileName = PhraseImpl.getLocaleHash(distId, envId, localeCode);
      final arb = loadArb('test.arb');

      setUp(() {
        when(diskCacheMock.read(any)).thenAnswer((realInvocation) async => arb);
      });

      test('loads translations from cache', () async {
        final phraseImpl = PhraseImpl(
          distId,
          envId,
          false,
          null,
          PhraseHost.eu,
          diskCache: diskCacheMock,
          prefs: prefsMock,
          api: apiMock,
          chef: chefMock,
        );

        await phraseImpl.onLocale(locale);

        verify(diskCacheMock.read(fileName));
        verifyNoMoreInteractions(diskCacheMock);
        verifyZeroInteractions(prefsMock);
        verifyZeroInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });

      test('loads translations from api', () async {
        final uuid = mockUUID();
        const version = '1';
        final lastUpdate = DateTime.now().toIso8601String();
        final apiResult = PhraseApiResult.success(version, arb);
        final phraseImpl = PhraseImpl(
          distId,
          envId,
          true,
          null,
          PhraseHost.eu,
          diskCache: diskCacheMock,
          prefs: prefsMock,
          api: apiMock,
          chef: chefMock,
        );
        when(prefsMock.getUUID()).thenAnswer((realInvocation) async => uuid);
        when(prefsMock.getVersion(any))
            .thenAnswer((realInvocation) async => version);
        when(prefsMock.getLastUpdate())
            .thenAnswer((realInvocation) async => lastUpdate);
        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);

        await phraseImpl.onLocale(locale);

        verify(diskCacheMock.read(fileName));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.eu,
          localeCode,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
        ));
      });
    });
  });
}
